/** skeletons.c
 *
 *  generic functions for dataparallel skeletons 
 *  map, reduce, prefix
 *
 *  C. Kessler reworked 991202
 */
#include <fork.h>   // always required in Fork programs
#include <io.h>     // printf() prototype 
#include <string.h> // memcpy() prototype
#include <assert.h>
#include <skeletons.h>


// generic non-nestable map routine for unary functions:

sync void map( sh straight void (*f)(void *, void *), sh void **x,
               sh void **d, sh int N, sh int elsize )
{
  int i;
  forall (i, 0, N, # )
     f( x + i*elsize, d + i*elsize );
}


// generic non-nestable map routine for binary functions with 1 scalar opnd:

sync void map1( sh straight void (*f)(void *, void *, void *),
                sh void **x, sh void **d, sh void **d2,
                sh int N, sh int elsize )
{
  int i;
  forall (i, 0, N, # )
     f( x + i*elsize, d + i*elsize, d2 );
}


// generic non-nestable map routine for binary functions:

sync void map2( sh straight void (*f)(void *, void *, void *), 
                sh void **x, sh void **d, sh void **d2,
                sh int N, sh int elsize )
{
  int i;
  forall (i, 0, N, # )
     f( x + i*elsize, d + i*elsize, d2 + i*elsize );
}


// #define max(a,b) ((a)>=(b)?(a):(b))  // defined in math.h

// generic nestable map routine for unary functions:

sync void Map( sync sh void (*f)(sh void *, sh void *, sh int), 
               sh void **x, sh void **d, 
               sh int N, sh int mx, sh int md, sh int elsize )
{
  sh int m, t, p = 0;
  int i;
  farm m = max(mx,md);   // extent of inner calculation
  farm pprintf("call Map(%d,%d,%d)\n", N, m, elsize);

  $ = mpadd(&p,1); // assert consecutive numbering of $ in 0..p-1

  if (p < N)
    for ( t=0; t<N; t+=p )
      fork(p; @=$; $=0) 
        if (t+@ < N)
          f( x + (t+@)*m*elsize, d + (t+@)*m*elsize, m );
  else 
    fork( N; @=$%N; $=$/N )
      f( x + @*m*elsize, d + @*m*elsize, m );
}





// nestable map skeleton for binary functions
// Set md2 to zero if d2 denotes a scalar value.

sync void Map2( sync sh void (*f)(sh void *, sh void *, sh void *, sh int),
                sh void **x, sh void **d, sh void **d2, 
                sh int N, sh int mx, sh int md, sh int md2, sh int elsize )
{
  sh int m;
  sh int p = #;
  farm m = max(mx,max(md,md2));   // extent of inner calculation
  farm pprintf("call Map2(N=%d,m=%d,elsize=%d) $=%d\n", N, m, elsize, $);
  if (p < N)
    fork(p; @=$$; ) {
      sh int t = @;
      farm pprintf("vor for: P%d: &t=%x t=%d @=%d $=%d\n",
            __PROC_NR__, &t, t, @, $ );
      for (/*t=@*/; t<N; t+=p) {
        farm pprintf("P%d: t=%d N=%d @=%d $=%d\n",
                      __PROC_NR__, t, N, @, $ );
        f( x + t*mx*elsize, d + t*md*elsize, d2 + t*md2*elsize, m);
      }
      farm pprintf("nach for: P%d: &t=%x t=%d N=%d @=%d $=%d gs=%d\n",
            __PROC_NR__, &t, t, N, @, $, groupsize() );
    }
  else 
    fork( N; @=$$%N; )
      f( x + @*mx*elsize, d + @*md*elsize, d2 + @*md2*elsize, m);
}


/** In-place f-reduction for an array d of n<=2p items of size elsize.
 *  Executed by p processors numbered consecutively from 0 to p-1.
 */
sync void preduce( sh straight void (*f)(void *, void *, void *),
                   sh void **d, sh int n, sh int elsize )
{
 sh int t;
 //seq if (n <= 2*groupsize())
 //   printf("Error: n=%d p=%d\n", n, groupsize());
 // iterative computation upwards the tree:
 for (t=1; t<n; t = t<<1 )   // sequential loop over tree levels
   if (2*$*t+t < n) 
      f( d+(2*$*t)*elsize, d+(2*$*t)*elsize, d + (2*$*t+t)*elsize );
}


void seq_reduce( straight void (*f)(void *, void *, void *),
                 void *s, void **d, int n, int elsize )
{
 int t;
 if (n<=0) return;
 memcpy( s, d, elsize );    // initialize s by d[0]
 for (t=1; t<n; t++)
    f( s, s, d+t*elsize );
}


sync void reduce( straight void (*f)(void *, void *, void *),
                  sh void *s, sh void **d, sh int n, sh int elsize )
{
 sh int p = 0; 
 sh void **temp;
 int ss, myslice;
 sh int t;
 $ = mpadd(&p,1); // assert consecutive numbering of $ in 0..p-1
 farm pprintf("call reduce(%d)\n", n );
 
 temp = (void **)shalloc( 2*p*elsize );  // temporary array

 if (2*p < n) {
   // partition the data vector d into n/p slices
   farm {
     ss = (int)((float)(n+p-1) / (float)(p));
     pprintf("ss=%d\n", ss);
     if ($*ss >= n)          myslice = 0;
     else if (($+1)*ss > n)  myslice = n - $*ss;
     else                    myslice = ss;
     pprintf("myslice=%d\n", myslice);
   }
   // and concurrently do sequential reduction for each slice:
   farm
     seq_reduce( f, temp+$*elsize, d+$*ss*elsize, myslice, elsize );
   n = p;  // extent of preduce
 }
 else 
   // copy <=2p data items in parallel to temp array:
   farm {
     if ($<n)
       memcpy( temp+$*elsize, d+$*elsize, elsize );
     if (p+$ < n)
       memcpy( temp+(p+$)*elsize, d+(p+$)*elsize, elsize );
   }

 preduce( f, temp, n, elsize );

 seq
   memcpy( s, temp, elsize );    // write result to *s
 shallfree();  // release temp
}



/** In-place iterative parallel f-prefix computation
 *  for a p-element array with p processors.
 *  If non--null, offset is f-combined with every element.
 *  Consecutive numbering of $ in 0..groupsize()-1 is assumed.
 */
sync void pprefix( sync sh void (*f)(void *, void *, void *),
                    sh void **x, sh void *offset, sh int elsize )
{
 sh int p = groupsize(); 
 sh int t;
  
 if (offset)
   if ($==0)
     f( x, offset, x );

 // iterative computation upwards the tree:
 for (t=1; t<p; t = t<<1 )  // sequential loop over tree levels
   if ($ >= t) 
      f( x+$*elsize, x+($-t)*elsize, x+$*elsize );

}

sync void prefix( sync sh void (*f)(void *, void *, void *),
                  sh void **x, sh void **d,
                  sh int n, sh int elsize )
{
 int i;
 sh int t;
 sh void *offset = NULL;
 sh int iter, p = 0; 
 $ = mpadd(&p,1); // assert consecutive numbering of $ in 0..p-1
  
 if ($ >= n) return;   // need at most n processors

 farm
   FORALL( i, &iter, 0, n, 1 )
     memcpy( x+i*elsize, d+i*elsize, elsize );    // x[i]=d[i]

 // outer sequential loop over x in slices of size p:
 for (t=0; t<n; t += p ) {
   if ($ < n-t)    // do not exceed bound of x for last slice
     pprefix( f, x+t*elsize, offset, elsize );
   offset = x + (t+p-1)*elsize; // offset for next slice prefix
 }
}
